import numpy as np
import matplotlib.pyplot as plt
import skimage
from skimage.feature import hog
from skimage import exposure

#HOG
img = plt.imread('sign1.png')
img = plt.imread('braided_0081.jpg')

#%%

def hog_fct(img):
    img = skimage.transform.resize(img, [128,128])
    hog_desc, hog_image = hog(img, orientations=9, pixels_per_cell=(8, 8), cells_per_block=(2, 2), visualize=True, channel_axis=-1)
    return hog_desc, hog_image

hog_desc, hog_image = hog_fct(img)

fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(8, 4)) #sharex=True, sharey=True)

ax1.axis('off')
ax1.imshow(img)
ax1.set_title('Input image')

# Rescale histogram for better display
hog_image_rescaled = exposure.rescale_intensity(hog_image, in_range=(0, 10))

ax2.axis('off')
ax2.imshow(hog_image_rescaled, cmap=plt.cm.gray)
ax2.set_title('Histogram of Oriented Gradients')
plt.show()

#%% LBP

def LBP(img):

    if img.ndim==3:
        img = skimage.color.rgb2gray(img)    
    h, w = img.shape
    LBP_map = np.zeros((h,w))
    weight = np.array([[1,2,4],[8,0,16],[32,64,128]])
    
    for i in range(1,h-1):
        for j in range(1,w-1):
            win = img[i-1:i+2,j-1:j+2]
            win_ = (win - img[i,j])>=0
            win_weight = weight*win_
            LBP_map[i,j] = np.sum(win_weight)
    LBP_histo, tmp = np.histogram(LBP_map, 256)
    return LBP_map, LBP_histo

LBP_map, LBP_histo = LBP(img)
plt.figure()
plt.subplot(121)
plt.imshow(img, cmap='gray'); plt.title('Image')
plt.show()
plt.subplot(122)
plt.imshow(LBP_map, cmap='gray')

plt.figure()
plt.hist(LBP_map.ravel(), 128)
plt.show()
plt.title('LBP histogram 128-bin')

plt.figure()
plt.hist(LBP_map.ravel(), 256)
plt.show()
plt.title('LBP histogram 256-bin')


#%% Block-wise LBP

def LBP_BW(img, cell_size = 16):
    if img.ndim==3:
        img = skimage.color.rgb2gray(img)    
    h, w = img.shape
    LBP_map, LBP_histo = LBP(img)
    LBP_histo_BW = []
    for i in range(0,h-cell_size+1,cell_size):
        for j in range(0,w-cell_size+1,cell_size):
            win = LBP_map[i:i+cell_size,j:j+cell_size]
            lbp_histo, tmp = np.histogram(win, 256)
            LBP_histo_BW.append(lbp_histo)
    LBP_histo_BW = np.array(LBP_histo_BW)
    return LBP_histo_BW

LBP_histo_BW = LBP_BW(img).ravel()


#%% Compute for all images

IL = np.load('I_train.npy')
IT = np.load('I_test.npy')
YL = np.load('Y_train.npy')
YT = np.load('Y_test.npy')

#%% Feature size

feature = 2 #1, 2, 3
img = IL[:,:,:,0]

if (feature==1):
    lbp_map, lbp_histo = LBP(img)
    p = lbp_histo.shape[0]
    XL = np.load('XL_lbp.npy')
    XT = np.load('XT_lbp.npy')
if (feature==2):
    lbp_histo_BW = LBP_BW(img).ravel()
    p = lbp_histo_BW.shape[0]
    XL = np.load('XL_lbp_bw.npy')
    XT = np.load('XT_lbp_bw.npy')
if (feature==3):
    hog_desc, hog_image = hog_fct(img)
    p = hog_desc.shape[0]
    XL = np.load('XL_hog.npy')
    XT = np.load('XT_hog.npy')


#%% Recompute descriptors
XL = np.zeros((IL.shape[-1], p))
for i in range(0,IL.shape[-1]):
    print(i)
    img = IL[:,:,:,i]
    if (feature == 1):
        lbp_map, desc = LBP(img)
    if (feature == 2):
        desc = LBP_BW(img)
    if (feature == 3):
        desc, tmp = hog_fct(img)
    XL[i,:] = desc.ravel()

XT = np.zeros((IT.shape[-1], p))
for i in range(0,IT.shape[-1]):
    print(i)
    img = IT[:,:,:,i]
    if (feature == 1):
        lbp_map, desc = LBP(img)
    if (feature == 2):
        desc = LBP_BW(img)
    if (feature == 3):
        desc, tmp = hog_fct(img)
    XT[i,:] = desc.ravel()


#%% SVM classifier
import time
from sklearn.svm import SVC

classifier = SVC(C = 100, kernel="rbf",  random_state=0)
start = time.time()
classifier.fit(XL, YL.ravel())
end = time.time()
print(end - start)

YT_pred = classifier.predict(XT)
accuracy_SVM = np.mean(YT_pred==YT.ravel())
print(f"{accuracy_SVM = }")



#LBP    = 0.55
#LBP_BW = 0.32
#HOG    = 0.35

#%% PCA
from sklearn.decomposition import PCA

pca = PCA(n_components=50, svd_solver='full') #when data too large, svd solver = randomize
pca.fit(XL)
XL_pca = pca.transform(XL)
XT_pca = pca.transform(XT)

print("ok")

#%% 
classifier = SVC(C = 100, kernel="rbf",  random_state=0)
start = time.time()
classifier.fit(XL_pca, YL.ravel())
end = time.time()
print(end - start)
YT_pred_pca = classifier.predict(XT_pca)
accuracy_SVM_PCA = np.mean(YT_pred_pca==YT.ravel())
print(f"{accuracy_SVM_PCA = }")

#LBP 50    = 0.58
#LBP_BW 50 = 0.45
#HOG 50    = 0.43


#%%

np.save('XL_lbp.npy', XL)
np.save('XT_lbp.npy', XT)

